% This demonstration program implements the algorithm published in the
% article entitled:
% 'Notably improved inversion of Differential Mobility
% Particle Sizer data obtained under conditions of fluctuating particle
% number concentrations'
%by B. M�lgaard, J. Vanhatalo, P.P. Aalto, N.L. Prisle, and K. H�meri.
% See the supplement cover page provided with this algorithm for full
% article details. Please, refer to this article if you use this algorithm.


% In order to run the demo you need to 
%   * first install GPstuff v-4.5 package
%     http://becs.aalto.fi/en/research/bayes/gpstuff/
%   * run the demo in the same folder where the demo files are or add
%     the folder to Matlab search path

% Copyright (c) 2015 Jarno Vanhatalo and Bjarke Molgaard

% This software is distributed under the GNU General Public
% License (version 3 or later); please refer to the file
% License.txt, included with the software, for details.

% ===== set up data structures =====
% Load example data
data = load('testdata20150302');%see data_format.txt
date='150302';
%Choose scans for the inversion
Nscan = 9;      %Number of scans to include
firstScan=64;   % all scans before this number are ignored

% Transfer matrix. Choose the part related to the chosen scans.
Tindex = data.Tindex;
T= data.T(Tindex(firstScan,1)+1:Tindex(firstScan+Nscan,1),...
    Tindex(firstScan,2)+1:Tindex(firstScan+Nscan,2));

% Particle counts (y in the article)
y = [];
for iScan=firstScan-1 +(1:Nscan)
    y = [y; data.count1(iScan,:)' ; data.count2(iScan,:)'];   % "outputs"
end

% Times of measurements (t in the article) and log10 particle sizes (u in
% the article) corresponding to the particle counts
x=[data.timevector(Tindex(firstScan,1)+1:Tindex(firstScan+Nscan,1)) ...
    log10(data.sizevector(Tindex(firstScan,1)+1:Tindex(firstScan+Nscan,1)))];

% Volume(s) of measurements (V in the article)
z = [];
for iScan=firstScan-1 +(1:Nscan)
    z = [z; data.volume1(iScan,:)' ; data.volume2(iScan,:)']; % the volumes
end
% ===== END: set up data structures =====


% ===== define the model, create model structures, and run inference =====
% Create likelihood and covariance function
lik = lik_countNumber('T', T); 

% Covariance structures and priors for their hyperparameters
% * You can test also different covariances. See the GPstuff manual for
%   different possibilities and how to use them:
%   http://arxiv.org/abs/1206.5754

% covariance structure corresponding to k(u,u') and priors for its
% hyperparameters
gpcf_p = gpcf_matern52('magnSigma2', 10, 'lengthScale', 1, ...
    'magnSigma2_prior', prior_t('s2',10) , 'lengthScale_prior', prior_t('s2',0.25), 'selectedVariables', 2);
% covariance structure corresponding to k(t,t') and priors for its
% hyperparameters
gpcf_t = gpcf_exp('magnSigma2', 1, 'lengthScale', 0.03, ...
    'magnSigma2_prior', prior_fixed , 'lengthScale_prior', prior_t('s2',0.01), 'selectedVariables', 1);
% covariance structure corresponding to k(u,u')*k(t,t')
gpcf_pt = gpcf_prod('cf', {gpcf_p, gpcf_t});
% covariance structure corresponding to mean \mu ~ N(0, \sigma^2)
gpcfc = gpcf_constant('constSigma2', 10, 'constSigma2_prior', prior_fixed);

% Define the options for the 'stabilized-newton' algorithm
latent_opt.tol = 1e-8;                                  % f optimization tolerance
latent_opt.maxiter = 450;                               % maximum number of iterations in stabilized Newton algorithm

% The Gaussian process (GP) model structure containing the likelihood and
% covariance functions and information concerning inference
%  * jitterSigma2:  Add "jitter" to the diagonal of the covariance function
gp = gp_setDMPS('lik', lik, 'cf', {gpcf_pt gpcfc}, 'latent_opt', latent_opt, 'jitterSigma2', 1e-4);

% Find the maximum a Posterior approximation of the hyperparameters (\theta) and
% construct Laplace approximation for f at the mode of \theta
opt=optimset('TolFun',1e-3,'TolX',1e-3,'Display','iter');  % Set optimization options into a struct
gp=gp_optimDMPS(gp,x,y,'z',z,'opt',opt);                   % optimize
% ===== END: model definition and inference ====

%% Visualize results

% the MAP of the hyperparameters
gp.cf{1}.cf{1}    % The covariance function along the particle size dimension
gp.cf{1}.cf{2}    % The covariance function along the time dimension

% Predict the particle number size distribution at 
% different time points

figure
ns=length(unique(x(:,2)));
timesteps=linspace(min(x(:,1)),max(x(:,1)),Nscan);
for i1=1:Nscan
    subplot(ceil(sqrt(Nscan)),ceil(sqrt(Nscan)),i1)
    xt = [repmat(timesteps(i1),ns,1) unique(x(:,2))];
    [Ef, Varf] = gpla_predDMPS(gp,x,y,xt,'z',z);
    semilogx(exp10(xt(:,2)),exp(Ef),'k')
    hold on
    semilogx(exp10(xt(:,2)),exp(Ef + 2*sqrt(Varf)),'k--')
    semilogx(exp10(xt(:,2)),exp(Ef - 2*sqrt(Varf)),'k--')
    ylabel('d{\itN}/dlog_1_0({\itD_p}) [cm^{-3}]')
    xlabel('{\itD_p} [m]');
%    semilogx(exp(xt(:,2)),exp(Ef),col)
%    hold on,
    %legend(sprintf('time = %.2f',timesteps(i1)))
    legend(datestr(timesteps(i1),13))
end


% Examine the predictions at data points
[Ef, Varf] = gpla_predDMPS(gp,x,y,x,'z',z);
figure
scatter(x(:,1),exp10(x(:,2)),3,Ef)
set(gca,'Yscale','log')
xlabel('Time')
ylabel('Particle diameter [m]')
colorbar
datetick('x')
title('Fitted values of f')


%Compare lambda with actual counts. The closer lambda and y are to each
%other, the better the model fit is. Note that there may be zero counts
%which are not seen because of the logarithmic scale.
figure
nv = size(T,2);
lambda=(exp(Ef)'*T)';
lambda = z.*lambda;
semilogy(reshape(lambda,nv,[]))
hold all, semilogy(reshape(y,nv,[]),'.')
legend('\lambda','y')
xlabel('Measurement index \it{i}')
ylabel('Number of particles counted/expected in each measurement')

% Examine predictions at a grid with one minute/ 5 s/ 10 s time-resolution
%hold on

step=0.1/log(10); %the resolution for the dp axis. Used also for the integration further down.
sizeaxis=log10(3e-9):step:log10(1e-6);
%timeaxis=min(x(:,1)):1/(24*60):max(x(:,1));%one minute
timeaxis=min(x(:,1)):1/(12*24*60):max(x(:,1));% 5 s
%timeaxis=min(x(:,1)):1/(6*24*60):max(x(:,1));% 10 s
xt=reshape(repmat(timeaxis,length(sizeaxis),1),[],1);
repeat=length(xt)/length(sizeaxis);
xt(:,2)=repmat(sizeaxis',repeat,1);

[Eft, Varft] = gpla_predDMPS(gp,x,y,xt,'z',z);

% scatter(xt(:,1),exp10(xt(:,2)),3,Eft)
% set(gca,'Yscale','log')
% xlabel('Time')
% ylabel('Particle diameter [m]')




figure
h=surface(timeaxis,exp10(sizeaxis),reshape(Eft,length(sizeaxis),[]))
%h=surface(timeaxis,exp(sizeaxis),reshape(Eft/log(10),length(sizeaxis),[]))
set(h,'Edgecolor','none')
set(gca,'Yscale','log')
xlabel('Time [hh:mm]')
datetick('x')
%hold all
%plot3(x(:,1),exp(x(:,2)),Ef,'.k')
hc=colorbar;
ylabel('{\itD_p} [m]')
ylim([3e-9 1e-6])
%caxis([0 5])
%set(hc,'Ytick',0:5,'Yticklabel',{'1','10','100','1000','10000','100000'})
%text(32.582,2e-8,'d{\itN}/dlog_1_0({\itD_p}) [cm^{-3}]','Rotation',90)
title('E(f)')

figure
h=surface(timeaxis,exp10(sizeaxis),reshape(Varft,length(sizeaxis),[]))
set(h,'Edgecolor','none')
set(gca,'Yscale','log')
xlabel('Time [hh:mm]')
datetick('x')
hold all
plot3(x(:,1),exp10(x(:,2)),Ef,'.k')
title('Var(f)')
colorbar
ylabel('{\itD_p} [m]')
ylim([3e-9 1e-6])


figure
h=surf(timeaxis,exp10(sizeaxis),reshape(exp(Eft),length(sizeaxis),[]))
set(gca,'Yscale','log')
set(h,'Edgecolor','none')
zlabel('d{\itN}/dlog_1_0({\itD_p}) [cm^{-3}]')
xlabel('Time [hh:mm]')
datetick('x')
ylabel('{\itD_p} [m]');

%----------------------------
%Compare particle number concentrations from new and old inversion algorithm
% and from CPC

%first load dmps data processed by the old inversion algorithm
dmpsOld=load(['dp' date '.sum']);
dpOld=dmpsOld(1,3:end);
dmpsOld(1,:)=[];

figure

dmps_sum=step*sum(exp(reshape(Eft+0.5*Varft,length(sizeaxis),[])));
plot(timeaxis,dmps_sum)
hold all
plot(reshape((repmat(dmpsOld(firstScan+(0:Nscan-1),1),1,3)+repmat([0 datenum(0,0,0,0,9,45) NaN],Nscan,1))',[],1),reshape([1;1;NaN]*dmpsOld(firstScan+(0:Nscan-1),2)',[],1))

%now load CPC data
cpc=load(['CH' date '.DA1'],'-ascii');
cpc(:,1)=floor(timeaxis(1))+cpc(:,1)/24; %changing time format
inTimeInterval=cpc(:,1)>timeaxis(1) & cpc(:,1)<timeaxis(end);

plot(cpc(inTimeInterval,1)-2.5/(24*3600),cpc(inTimeInterval,3))
%CPC time stamp are in the end of the averaging period. In the line above
%2.5 seconds were subtracted to get the middle of the averaging period.
legend('DMPS','DMPS old inversion','CPC')
xlabel('Time [hh:mm]')
datetick('x')
ylabel('PNC [cm^{-3}]')

%---------------------------
% sample from the posterior at time timeaxis(ii)
ii=100; %arbitrary choice. Change this index to choose the time you want.
xt=[ones(length(sizeaxis),1)*timeaxis(ii) sizeaxis']; 
fs = gpla_rndDMPS(gp,x,y,xt,'z',z, 'nsamp', 7);

figure
semilogx(exp10(sizeaxis),exp(fs))
ylabel('d{\itN}/dlog_1_0({\itD_p}) [cm^{-3}]')
xlabel('{\itD_p} [m]');
legend('Draw 1','Draw 2','Draw 3','Draw 4','Draw 5','Draw 6','Draw 7')
title(['Size distribution sampled from the posterior for time ' datestr(timeaxis(ii),13)])



fs = gpla_rndDMPS(gp,x,y,xt,'z',z, 'nsamp', 1e4);
dmps_sum_sample=step*sum(exp(fs));

