function [gp, varargout] = gp_optimDMPS(gp, x, y, varargin)
%GP_OPTIM  Optimize paramaters of a Gaussian process model for DMPS
%          inversion algorithm. Modified based on gp_optim in GPstuff.
%
%  Description
%    GP = GP_OPTIM(GP, X, Y, OPTIONS) optimises the parameters of a
%    GP structure given matrix X of training inputs and vector
%    Y of training targets.
%
%    [GP, OUTPUT1, OUTPUT2, ...] = GP_OPTIM(GP, X, Y, OPTIONS)
%    optionally returns outputs of the optimization function.
%
%    OPTIONS is optional parameter-value pair
%      z      - Volumes of scans
%      optimf - function handle for an optimization function, which is
%               assumed to have similar input and output arguments
%               as usual fmin*-functions. Default is @fminscg.
%      opt    - options structure for the minimization function.
%               Use optimset to set these options. By default options
%               'GradObj' is 'on', 'LargeScale' is 'off'.
%
%  See also
%    GP_SET, GP_E, GP_G, GP_EG, FMINSCG, FMINLBFGS, OPTIMSET, DEMO_REGRESSION*
%
%  References
%    Jarno Vanhatalo, Jaakko Riihimki, Jouni Hartikainen, Pasi Jylnki,
%    Ville Tolvanen and Aki Vehtari (2012). Bayesian Modeling with Gaussian
%    Processes using the GPstuff Toolbox. arXiv:1206.5754   

% Copyright (c) 2015 Jarno Vanhatalo
% Copyright (c) 2010-2012 Aki Vehtari

% This software is distributed under the GNU General Public
% License (version 3 or later); please refer to the file
% License.txt, included with the software, for details.

ip=inputParser;
ip.FunctionName = 'GP_OPTIM';
ip.addRequired('gp',@(x) isstruct(x) || isempty(x));
ip.addRequired('x', @(x) ~isempty(x) && isreal(x) && all(isfinite(x(:))))
ip.addRequired('y', @(x) ~isempty(x) && isreal(x) && all(isfinite(x(:))))
ip.addParameter('z', [], @(x) isreal(x) && all(isfinite(x(:))))
ip.addParameter('optimf', @fminscg, @(x) isa(x,'function_handle'))
ip.addParameter('opt', [], @isstruct)
ip.parse(gp, x, y, varargin{:});
z=ip.Results.z;
optimf=ip.Results.optimf;
opt=ip.Results.opt;

fh_eg=@(ww) gp_egDMPS(ww, gp, x, y, 'z', z);
optdefault=struct('GradObj','on','LargeScale','off');

opt=setOpt(optdefault,opt);
w=gp_pak(gp);
switch nargout
    case 6
        [w,fval,exitflag,output,grad,hessian] = optimf(fh_eg, w, opt);
        varargout={fval,exitflag,output,grad,hessian};
    case 5
        [w,fval,exitflag,output,grad] = optimf(fh_eg, w, opt);
        varargout={fval,exitflag,output,grad};
    case 4
        [w,fval,exitflag,output] = optimf(fh_eg, w, opt);
        varargout={fval,exitflag,output};
    case 3
        [w,fval,exitflag] = optimf(fh_eg, w, opt);
        varargout={fval,exitflag};
    case 2
        [w,fval] = optimf(fh_eg, w, opt);
        varargout={fval};
    case 1
        w = optimf(fh_eg, w, opt);
        varargout={};
end
gp=gp_unpak(gp,w);
end

function opt=setOpt(optdefault, opt)
  % Set default options
  opttmp=optimset(optdefault,opt);
  
  % Set some additional options for @fminscg
  if isfield(opt,'lambda')
    opttmp.lambda=opt.lambda;
  end
  if isfield(opt,'lambdalim')
    opttmp.lambdalim=opt.lambdalim;
  end
  opt=opttmp;
end

function [e, g] = gp_egDMPS(w, gp, x, y, varargin)
e=gpla_eDMPS(w, gp, x, y, varargin{:});
if nargout>1
    if isnan(e)
        g=NaN;
    else
        g=gpla_gDMPS(w, gp, x, y, varargin{:});
    end
end
end