function gp = gp_setDMPS(varargin)
%GP_SET  Create a Gaussian process model structure for DMPS inversion
%        algorithm. This is modified from gp_set of GPstuff v4.5
%
%  Description
%    GP = GP_SET('PARAM1',VALUE1,'PARAM2',VALUE2,...)
%    creates a Gaussian process structure in which the named
%    parameters have the specified values. Any unspecified
%    parameters are set to default values. 'cf' parameter has to be
%    specified.  
%  
%    GP = GP_SET(GP,'PARAM1',VALUE1,'PARAM2',VALUE2,...) 
%    modify a Gaussian process structure with the named
%    parameters altered with the specified values.
%
%    Parameters for Gaussian process
%      cf           - Single covariance structure or cell array of 
%                     covariance function structures created by
%                     gpcf_* functions. The default is [],
%                     except if neither cf or meanf is given then the
%                     the default is {gpcf_constant() gpcf_sexp()}. 
%      lik          - Likelihood structure created by one of the 
%                     likelihood functions lik_*. The default is
%                     created by lik_gaussian(). If likelihood is
%                     non-Gaussian, see latent_method below.
%      jitterSigma2 - Positive jitter to be added in the diagonal of 
%                     covariance matrix. The default is 1e-9.
%
%    The additional fields are:
%      latent_opt    - Additional option structure for the chosen
%                      latent method (Laplace). See default values for
%                      options below.
%        tol          - Iterations are stopped when change in the log 
%                       marginal likelihood is smaller than tol. The 
%                       default is 1e-6.
%
%  References
%    Jarno Vanhatalo, Jaakko Riihimki, Jouni Hartikainen, Pasi Jylnki,
%    Ville Tolvanen and Aki Vehtari (2012). Bayesian Modeling with Gaussian
%    Processes using the GPstuff Toolbox. arXiv:1206.5754   
%
%    B. Mlgaard, J. Vanhatalo, P. Aalto, N.L. Prisle, and K. Hmeri
%    (manuscript). Notably improved inversion of Differential Mobility
%    Particle Sizer data obtained under conditions of fluctuating particle
%    number concentrations  
%
%  See also
%    GPCF_*, LIK_*, PRIOR_*, GP_PAK, GP_UNPAK, GP_E, GP_G, GP_EG,
%    GP_PRED, GP_MC, GP_IA, ...


% Copyright (c) 2006-2010, 2015 Jarno Vanhatalo
% Copyright (c) 2010-2011 Aki Vehtari
% Copyright (c) 2014 Arno Solin and Jukka Koskenranta
  
% This software is distributed under the GNU General Public 
% License (version 3 or later); please refer to the file 
% License.txt, included with the software, for details.

  ip=inputParser;
  ip.FunctionName = 'GP_SET';
  ip.addOptional('gp', [], @isstruct);
  ip.addParameter('cf',[], @(x) isempty(x) || isstruct(x) || iscell(x));
  ip.addParameter('lik',lik_gaussian(), @(x) isstruct(x));
  ip.addParameter('jitterSigma2',1e-9, @(x) isscalar(x) && x>=0);
  ip.addParameter('latent_opt',struct(), @isstruct);
  ip.parse(varargin{:});
  gp=ip.Results.gp;


  if isempty(gp)
    % Initialize a Gaussian process
    init=true;
  else
    % Modify a Gaussian process
    if ~isfield(gp,'cf')
      error('First argument does not seem to be a Gaussian process structure')
    end
    init=false;
  end

  gp.type='FULL';             % type of GP (a regular GP without sparse approximations)
  gp.infer_params= 'covariance+likelihood';   % infer all hyperparameters
    
  % Likelihood
  if init || ~ismember('lik',ip.UsingDefaults)
    gp.lik = ip.Results.lik;
  end
  % Covariance function(s)
  if init || ~ismember('cf',ip.UsingDefaults)
    gp.cf=ip.Results.cf;
    if isstruct(gp.cf)
      % store single structure in a cell array, too
      gp.cf={gp.cf};
    end
  end
  % Jitter
  if init || ~ismember('jitterSigma2',ip.UsingDefaults)
    gp.jitterSigma2=ip.Results.jitterSigma2;
  end
  
  % Set latent method
  gp.latent_method='Laplace';
  % following sets gp.fh.e = @laplace_algorithm;
  gp = gpla_eDMPS('init', gp);

  % Options for Laplace approximation
  if init || ~ismember('latent_opt',ip.UsingDefaults) || ~isfield(gp,'latent_opt')
      latent_opt=ip.Results.latent_opt;
      % Handle latent_opt
      ipla=inputParser;
      ipla.FunctionName = 'GP_SET - latent method Laplace options';
      ipla.addParameter('tol',1e-6, @(x) isreal(x) && isscalar(x) && isfinite(x) && x>0);
      ipla.addParameter('maxiter', 100, @(x) isreal(x) && isscalar(x) && isfinite(x) && x>0);
      ipla.parse(latent_opt);
      if init || ~ismember('tol',ipla.UsingDefaults) || ~isfield(gp.latent_opt,'tol')
          gp.latent_opt.tol = ipla.Results.tol;
      end
      if init || ~ismember('maxiter',ipla.UsingDefaults) || ~isfield(gp,'maxiter')
          gp.latent_opt.maxiter = ipla.Results.maxiter;
      end
  end 
  
end
