function [e, edata, eprior, param] = gpla_eDMPS(w, gp, varargin)
%GPLA_E  Construct Laplace approximation and return marginal log posterior
%        estimate for the DMPS inversion algorithm. 
%        (code is based on gpla_e function in GPstuff package)
%
%  Description
%    E = GPLA_E(W, GP, X, Y, OPTIONS) takes a GP structure GP
%    together with a matrix X of input vectors and a matrix Y of
%    target vectors, and finds the Laplace approximation for the
%    conditional posterior p(Y | X, th), where th is the
%    parameters. Returns the energy at th (see below). Each
%    row of X corresponds to one input vector and each row of Y
%    corresponds to one target vector.
%
%    [E, EDATA, EPRIOR] = GPLA_E(W, GP, X, Y, OPTIONS) returns also
%    the data and prior components of the total energy.
%
%    The energy is minus log posterior cost function for th:
%      E = EDATA + EPRIOR
%        = - log p(Y|X, th) - log p(th),
%      where th represents the parameters (lengthScale,
%      magnSigma2...), X is inputs and Y is observations.
%
%    OPTIONS is optional parameter-value pair
%      z - volumes of scans
%
%  References
%
%    Rasmussen, C. E. and Williams, C. K. I. (2006). Gaussian
%    Processes for Machine Learning. The MIT Press.
%
%   B. Mlgaard, J. Vanhatalo, P. Aalto, N.L. Prisle, and K. Hmeri
%   (manuscript). Notably improved inversion of Differential Mobility
%   Particle Sizer data obtained under conditions of fluctuating particle
%   number concentrations  
%
%  See also
%    GP_SET, GP_E, GPLA_GDMPS, GPLA_PREDDMPS

%  Description 2
%    Additional properties meant only for internal use.
%
%    GP = GPLA_E('init', GP) takes a GP structure GP and
%    initializes required fields for the Laplace approximation.
%
%    GP = GPLA_E('clearcache', GP) takes a GP structure GP and clears the
%    internal cache stored in the nested function workspace.

% Copyright (c) 2015 Bjarke Molgaard
% Copyright (c) 2007-2010, 2015 Jarno Vanhatalo
% Copyright (c) 2010 Aki Vehtari
% Copyright (c) 2010 Pasi Jylnki

% This software is distributed under the GNU General Public
% License (version 3 or later); please refer to the file
% License.txt, included with the software, for details.

% parse inputs
ip=inputParser;
ip.FunctionName = 'GPLA_EDMPS';
ip.addRequired('w', @(x) ...
    isempty(x) || ...
    (ischar(x) && strcmp(w, 'init')) || ...
    isvector(x) && isreal(x) && all(isfinite(x)) ...
    || all(isnan(x)));
ip.addRequired('gp',@isstruct);
ip.addOptional('x', @(x) isnumeric(x) && isreal(x) && all(isfinite(x(:))))
ip.addOptional('y', @(x) isnumeric(x) && isreal(x) && all(isfinite(x(:))))
ip.addParameter('z', [], @(x) isnumeric(x) && isreal(x) && all(isfinite(x(:))))
ip.parse(w, gp, varargin{:});
x=ip.Results.x;
y=ip.Results.y;
z=ip.Results.z;

if strcmp(w, 'init')
    % Initialize cache
    ch = [];
    
    % set function handle to the nested function laplace_algorithm
    % this way each gp has its own peristent memory for Laplace
    % approximation which speeds up the computation
    gp.fh.ne = @laplace_algorithm;
    % set other function handles
    gp.fh.e=@gpla_eDMPS;
    gp.fh.g=@gpla_gDMPS;
    gp.fh.pred=@gpla_predDMPS;
    e = gp;
    % remove clutter from the nested workspace
    clear w gp varargin ip x y z
elseif strcmp(w, 'clearcache')
    % clear the cache
    gp.fh.ne('clearcache');
else
    % call laplace_algorithm using the function handle to the nested function
    % this way each gp has its own peristent memory for Laplace which
    % speeds up the computation 
    [e, edata, eprior, param] = gp.fh.ne(w, gp, x, y, z);
end

    function [e, edata, eprior, param] = laplace_algorithm(w, gp, x, y, z)
        
        if strcmp(w, 'clearcache')
            ch=[];
            return
        end
        % code for the Laplace algorithm
        
        % check whether saved values can be used. If yes return them,
        % otherwise run Laplace algorithm
        yt=y;
        zt=z;
        if size(x,1) ~= length(y)
            yt(end+1:size(x,1))=0;
            zt(end+1:size(x,1))=0;
        end
        datahash=hash_sha512_DMPS([x yt zt]);
        if ~isempty(ch) && all(size(w)==size(ch.w)) && all(abs(w-ch.w)<1e-8) && ...
                isequal(datahash,ch.datahash)
            % The covariance function parameters or data haven't changed so we
            % can return the energy and the Laplace algorithm that are
            % saved in the cache
            e = ch.e;
            edata = ch.edata;
            eprior = ch.eprior;
            param.f = ch.f;
            param.L = ch.L;
            param.La2 = ch.La2;
            param.a = ch.a;
        else
            % The parameters or data have changed since
            % the last call for gpla_eDMPS. In this case we need to
            % re-evaluate the Laplace approximation
            gp=gp_unpak(gp, w);
            ncf = length(gp.cf);            
            maxiter = gp.latent_opt.maxiter;
            tol = gp.latent_opt.tol;
            
            % Check that mean functions are not in use
            if isfield(gp,'meanf')
                error('meanf not implemented')
            end
            
            K = gp_trcov(gp, x);     % prior covariance of latent variables
            n=size(K,1);             % number of latent variables
            f = zeros(size(x,1),1);  % Initialize latent values to zero
            
            % --------------------------------------------------------------------------------
            % find the posterior mode of latent variables by stabilized Newton method.
            % --------------------------------------------------------------------------------            
            a=f;            
            LK = chol(K, 'lower');
            lp = -f'*(LK'\(LK\f))/2 +gp.lik.fh.ll(gp.lik, y, f, z);
            
            lp_old = -Inf;
            f_old = f+1;
            W = -gp.lik.fh.llg2(gp.lik, y, f, 'latent', z);
            dlp = gp.lik.fh.llg(gp.lik, y, f, 'latent', z) - LK'\(LK\f);
            
            iter=0;
            % begin Newton's iterations
            while (lp - lp_old > tol || max(abs(f-f_old)) > tol) && iter < maxiter
                iter=iter+1;
                lp_old = lp; a_old = a; f_old = f;
                
                B = LK'*W*LK;
                B(1:n+1:end)=B(1:n+1:end)+1;
                [LB, notpositivedefinite] = chol(B, 'lower');
                
                if notpositivedefinite    % stabilized-newton iteration
                    sW = full(sqrt(max(diag(W),0)));
                    LB=bsxfun(@times,bsxfun(@times,sW,K),sW');
                    LB(1:n+1:end)=LB(1:n+1:end)+1;
                    LB = chol(LB, 'lower');
                    b = K*dlp;
                    delta = b - K*(sW.*(LB'\(LB\(sW.*b))));
                else                        % Regular Newton iteration                    
                    delta = LK*(LB'\(LB\(LK'*dlp)));
                end
                f = f + delta;
                if any(isnan(f))
                    [edata,e,eprior,param,ch] = set_output_for_notpositivedefinite();
                    return
                end
                a = LK'\(LK\f);
                lp_new = -a'*f/2 + gp.lik.fh.ll(gp.lik, y, f, z);
                i = 0;
                while i < 25 && (lp_new < lp_old  || isnan(sum(f)) || isnan(lp_new))
                    % reduce step size by half until improvement in
                    % objective
                    a = (a_old+a)/2;
                    f = K*a;
                    lp_new = -a'*f/2 + gp.lik.fh.ll(gp.lik, y, f, z);
                    i = i+1;
                end
                %if i>=25
                %    warning('step size might still be too large in Newton iteration')
                %end
                lp = lp_new;
                W = -gp.lik.fh.llg2(gp.lik, y, f, 'latent', z);
                dlp = gp.lik.fh.llg(gp.lik, y, f, 'latent', z) - LK'\(LK\f);
            end
            if iter>=maxiter
                warning('Newton iteration may not have converged. You might want to increase maxiter.')
            end
            
            % Calculate the approximate log marginal likelihood            
            edata = -lp + sum(log(diag(LB)));
            
            if notpositivedefinite
                [edata,e,eprior,param,ch] = set_output_for_notpositivedefinite();
                return
            end
            
            % ======================================================================
            % Evaluate the prior contribution to the error from covariance functions
            % ======================================================================
            eprior = 0;
            for i1=1:ncf
                gpcf = gp.cf{i1};
                eprior = eprior - gpcf.fh.lp(gpcf);
            end
            
            % ======================================================================
            % Evaluate the prior contribution to the error from likelihood function
            % ======================================================================
            if isfield(gp, 'lik') && isfield(gp.lik, 'p')
                lik = gp.lik;
                eprior = eprior - lik.fh.lp(lik);
            end
            
            % ============================================================
            % Evaluate the prior contribution to the error from mean functions
            % ============================================================
            if ~isempty(strfind(gp.infer_params, 'mean'))
                for i=1:length(gp.meanf)
                    gpmf = gp.meanf{i};
                    eprior = eprior - gpmf.fh.lp(gpmf);
                end
            end
            
            e = edata + eprior;
            
            % store values to struct param
            param.f = f;
            param.La2 = W;
            %param.L = chol(K);
            param.L = LK;
            param.a = a;
            
            % store values to the cache
            ch = param;
            ch.w = w;
            ch.e = e;
            ch.edata = edata;
            ch.eprior = eprior;
            ch.n = size(x,1);
            ch.datahash=datahash;
        end
    end

    function [edata,e,eprior,param,ch] = set_output_for_notpositivedefinite()
        % Instead of stopping to chol error, return NaN
        edata=NaN;
        e=NaN;
        eprior=NaN;
        param.f=NaN;
        param.L=NaN;
        param.a=NaN;
        param.La2=NaN;
        param.p=NaN;
        w=NaN;
        datahash = NaN;
        ch=param;
        ch.e = e;
        ch.edata = edata;
        ch.eprior = eprior;
        ch.datahash=datahash;
        ch.w = NaN;
    end

end
