function [g, gdata, gprior] = gpla_gDMPS(w, gp, x, y, varargin)
%GPLA_G   Evaluate gradient of Laplace approximation's marginal
%         log posterior estimate (GPLA_EDMPS) in the DMPS inversion
%         algorithm
%        (code is based on gpla_g function in GPstuff package)
%
%  Description
%    G = GPLA_GDMPS(W, GP, X, Y, OPTIONS) takes a full GP parameter
%    vector W, structure GP a matrix X of input vectors and a
%    matrix Y of target vectors, and evaluates the gradient G of
%    Laplace approximation's marginal log posterior estimate. Each row of X
%    corresponds to one input vector and each row of Y corresponds
%    to one target vector.
% 
%    [G, GDATA, GPRIOR] = GPLA_G(W, GP, X, Y, OPTIONS) also returns
%    the data and prior contributions to the gradient.
%
%    OPTIONS is optional parameter-value pair
%      z - volumes of scans
%
%  See also
%    GP_SET, GP_G, GPLA_EDMPS, GPLA_PREDDMPS
%
%  References
%    Rasmussen, C. E. and Williams, C. K. I. (2006). Gaussian
%    Processes for Machine Learning. The MIT Press.
%
%    B. Mlgaard, J. Vanhatalo, P. Aalto, N.L. Prisle, and K. Hmeri
%    (manuscript). Notably improved inversion of Differential Mobility
%    Particle Sizer data obtained under conditions of fluctuating particle
%    number concentrations  

% Copyright (c) 2007-2010,2015 Jarno Vanhatalo
% Copyright (c) 2010           Aki Vehtari

% This software is distributed under the GNU General Public
% License (version 3 or later); please refer to the file
% License.txt, included with the software, for details.

ip=inputParser;
ip.FunctionName = 'GPLA_GDMPS';
ip.addRequired('w', @(x) isvector(x) && isreal(x) && all(isfinite(x)));
ip.addRequired('gp',@isstruct);
ip.addRequired('x', @(x) ~isempty(x) && isreal(x) && all(isfinite(x(:))))
ip.addRequired('y', @(x) ~isempty(x) && isreal(x) && all(isfinite(x(:))))
ip.addParameter('z', [], @(x) isreal(x) && all(isfinite(x(:))))
ip.parse(w, gp, x, y, varargin{:});
z=ip.Results.z;

gp = gp_unpak(gp, w);       % unpak the parameters
[tmp,tmp,hier]=gp_pak(gp);   % Get the hierarchy of the parameters
ncf = length(gp.cf);

g = [];
gdata = [];
gprior = [];

if isfield(gp,'meanf')
    error('mean function not implemented')
end

% get the parameters of Laplace approximation
[e, ~, ~, p] = gpla_eDMPS(gp_pak(gp), gp, x, y, 'z', z);
[f, LK, a, W] = deal(p.f, p.L, p.a, p.La2);
if isnan(e)
    g=NaN; gdata=NaN; gprior=NaN;
    return;
end

%  The implementation follows the general results of Rasmussen and
%  Williams (2006) (page 125, http://www.gaussianprocess.org/gpml/chapters/RW5.pdf). 
%  The main difference is in how we calculate equation 5.23

% help matrices and vectors
V = LK'*W;
[LB, notpositivedefinite] = chol(eye(size(LK')) +  V*LK);
if notpositivedefinite
    % instead of stopping to chol error, return NaN
    g=NaN;
    gdata = NaN;
    gprior = NaN;
    return;
end
R2 = LB'\V;
R = W - R2'*R2;           % R = W - W*((inv(K) + W)\W) = inv( inv(W) +K );
RW = LB'\LK'; RW = RW'*RW;  % RW = inv(inv(K)+W);
DW = gp.lik.fh.llg3(gp.lik, y, f, 'latent', z) ;
s2 = zeros(size(f));
for h1=1:length(f)
    s2(h1) = 0.5*sum(sum( RW.*DW{h1} ));
end

% =================================================================
% Gradient with respect to the covariance function parameters
if ~isempty(strfind(gp.infer_params, 'covariance'))
    % Evaluate the gradients from covariance functions
    i1=0;
    for i=1:ncf
        gpcf = gp.cf{i};
        DKffc = gpcf.fh.cfg(gpcf, x); % derivative of covariance matrix
        np=length(DKffc);
        gprior_cf = -gpcf.fh.lpg(gpcf);
        
        g1 = gp.lik.fh.llg(gp.lik, y, f, 'latent', z);
        for i2 = 1:np
            DKff=DKffc{i2};
            i1 = i1+1;
            s1 = 0.5 * a'*DKff*a - 0.5*sum(sum(R.*DKff));
            b = DKff * g1;
            s3 = b - RW*W*b; %s3 = (eye(size(K)) + K*W)\b;
            gdata(i1) = -(s1 + s2'*s3);
        end        
        gprior=[gprior gprior_cf];
    end
    
end

% If the covariance function parameters of the model have additional
% hyperparameters that are not fixed set the gradients in correct order
g = gdata + gprior;

assert(isreal(gdata))
assert(isreal(gprior))
end
