function [Eft, Varft] = gpla_predDMPS(gp, x, y, varargin)
%GPLA_PRED  Predictions with Gaussian Process Laplace approximation
%           for the DMPS inversion algorithm. 
%           (code is based on gpla_pred function in GPstuff package)
%
%  Description
%    [EFT, VARFT] = GPLA_PRED(GP, X, Y, XT, OPTIONS)
%    takes a GP structure together with matrix X of training
%    inputs and vector Y of training targets, and evaluates the
%    predictive distribution at test inputs XT. Returns a posterior
%    mean EFT and variance VARFT of latent variables and the
%    posterior predictive mean EYT and variance VARYT.
%
%    OPTIONS is optional parameter-value pair
%      predcf - an index vector telling which covariance functions are
%               used for prediction. Default is all (1:gpcfn).
%               See additional information below.
%      z      - volumes of scans
%
%    NOTE! In case of FIC and PIC sparse approximation the
%    prediction for only some PREDCF covariance functions is just
%    an approximation since the covariance functions are coupled in
%    the approximation and are not strictly speaking additive
%    anymore.
%
%    For example, if you use covariance such as K = K1 + K2 your
%    predictions Eft1 = ep_pred(GP, X, Y, X, 'predcf', 1) and
%    Eft2 = ep_pred(gp, x, y, x, 'predcf', 2) should sum up to
%    Eft = ep_pred(gp, x, y, x). That is Eft = Eft1 + Eft2. With
%    FULL model this is true but with FIC and PIC this is true only
%    approximately. That is Eft \approx Eft1 + Eft2.
%
%    With CS+FIC the predictions are exact if the PREDCF covariance
%    functions are all in the FIC part or if they are CS
%    covariances.
%
%    NOTE! When making predictions with a subset of covariance
%    functions with FIC approximation the predictive variance can
%    in some cases be ill-behaved i.e. negative or unrealistically
%    small. This may happen because of the approximative nature of
%    the prediction.
%
%  References
%    Rasmussen, C. E. and Williams, C. K. I. (2006). Gaussian
%    Processes for Machine Learning. The MIT Press.
%
%    B. Mlgaard, J. Vanhatalo, P. Aalto, N.L. Prisle, and K. Hmeri
%    (manuscript). Notably improved inversion of Differential Mobility
%    Particle Sizer data obtained under conditions of fluctuating particle
%    number concentrations  
%
%  See also
%    GPLA_EDMPS, GPLA_GDMPS
%
% Copyright (c) 2007-2010, 2015 Jarno Vanhatalo
% Copyright (c) 2012            Aki Vehtari

% This software is distributed under the GNU General Public
% License (version 3 or later); please refer to the file
% License.txt, included with the software, for details.

ip=inputParser;
ip.FunctionName = 'GPLA_PRED';
ip.addRequired('gp', @isstruct);
ip.addRequired('x', @(x) ~isempty(x) && isreal(x) && all(isfinite(x(:))))
ip.addRequired('y', @(x) ~isempty(x) && isreal(x) && all(isfinite(x(:))))
ip.addOptional('xt', [], @(x) isempty(x) || (isreal(x) && all(isfinite(x(:)))))
ip.addParameter('z', [], @(x) isreal(x) && all(isfinite(x(:))))
ip.addParameter('predcf', [], @(x) isempty(x) || ...
    isvector(x) && isreal(x) && all(isfinite(x)&x>0))
if numel(varargin)==0 || isnumeric(varargin{1})
    % inputParser should handle this, but it doesn't
    ip.parse(gp, x, y, varargin{:});
else
    ip.parse(gp, x, y, [], varargin{:});
end
xt=ip.Results.xt;
z=ip.Results.z;
predcf=ip.Results.predcf;
if isempty(xt)
    xt=x;
end

% get the parameters of Laplace approximation
[~, ~, ~, p] = gpla_e(gp_pak(gp), gp, x, y, 'z', z);
[f, LK, W] = deal(p.f, p.L, p.La2);

% Calculate the covariance between training and prediction set
% notice the order xt,x to avoid transpose later
K_nf = gp_cov(gp,xt,x,predcf);

% Evaluate the mean
deriv = gp.lik.fh.llg(gp.lik, y, f, 'latent', z);
Eft = K_nf*deriv;

if nargout > 1
    % Evaluate the variance
    kstarstar = gp_trvar(gp,xt,predcf);
    V = LK'*W;
    R = chol(eye(size(LK)) +  V*LK);
    R2 = (R'\V)*K_nf';
    Varft = kstarstar - sum(K_nf.*(W*K_nf')',2) + sum(R2.*R2,1)';    
end
end
