function sampft = gpla_rndDMPS(gp, x, y, varargin)
%GPLA_RND  Random draws from the Gaussian Process Laplace approximation for
%          particle size distribution model
%
%  Description
%    SAMPFT = GPLA_RND(GP, X, Y, XT, OPTIONS)
%    takes a GP structure together with matrix X of training
%    inputs and vector Y of training targets, and evaluates the
%    predictive distribution at test inputs XT. Returns posterior samples
%    from the function f(xt) in matrix SAMPFT (size of size(x,1) x nsamp).
%
%    OPTIONS is optional parameter-value pair
%      nsamp     - determines the number of samples (default = 1).
%      predcf - an index vector telling which covariance functions are
%               used for prediction. Default is all (1:gpcfn).
%               See additional information below.
%      z      - optional observed quantity in triplet (x_i,y_i,z_i)
%               Some likelihoods may use this. For example, in case of
%               Poisson likelihood we have z_i=E_i, that is, expected value
%               for ith case.
%
%    NOTE! In case of FIC and PIC sparse approximation the
%    prediction for only some PREDCF covariance functions is just
%    an approximation since the covariance functions are coupled in
%    the approximation and are not strictly speaking additive
%    anymore.
%
%    For example, if you use covariance such as K = K1 + K2 your
%    predictions Eft1 = ep_pred(GP, X, Y, X, 'predcf', 1) and
%    Eft2 = ep_pred(gp, x, y, x, 'predcf', 2) should sum up to
%    Eft = ep_pred(gp, x, y, x). That is Eft = Eft1 + Eft2. With
%    FULL model this is true but with FIC and PIC this is true only
%    approximately. That is Eft \approx Eft1 + Eft2.
%
%    With CS+FIC the predictions are exact if the PREDCF covariance
%    functions are all in the FIC part or if they are CS
%    covariances.
%
%    NOTE! When making predictions with a subset of covariance
%    functions with FIC approximation the predictive variance can
%    in some cases be ill-behaved i.e. negative or unrealistically
%    small. This may happen because of the approximative nature of
%    the prediction.
%
%  References
%    Rasmussen, C. E. and Williams, C. K. I. (2006). Gaussian
%    Processes for Machine Learning. The MIT Press.
%
%    B. Mlgaard, J. Vanhatalo, P. Aalto, N.L. Prisle, and K. Hmeri
%    (manuscript). Notably improved inversion of Differential Mobility
%    Particle Sizer data obtained under conditions of fluctuating particle
%    number concentrations
%
%  See also
%    GPLA_E, GPLA_G, GP_PRED, DEMO_SPATIAL, DEMO_CLASSIFIC
%
% Copyright (c) 2007-2010, 2015 Jarno Vanhatalo
% Copyright (c) 2012 Aki Vehtari

% This software is distributed under the GNU General Public
% License (version 3 or later); please refer to the file
% License.txt, included with the software, for details.

ip=inputParser;
ip.FunctionName = 'GPLA_RNDmod';
ip.addRequired('gp', @isstruct);
ip.addRequired('x', @(x) ~isempty(x) && isreal(x) && all(isfinite(x(:))))
ip.addRequired('y', @(x) ~isempty(x) && isreal(x) && all(isfinite(x(:))))
ip.addOptional('xt', [], @(x) isempty(x) || (isreal(x) && all(isfinite(x(:)))))
ip.addParamValue('z', [], @(x) isreal(x) && all(isfinite(x(:))))
ip.addParamValue('predcf', [], @(x) isempty(x) || ...
    isvector(x) && isreal(x) && all(isfinite(x)&x>0))
ip.addParamValue('nsamp', 1, @(x) isreal(x) && isscalar(x))
if numel(varargin)==0 || isnumeric(varargin{1})
    % inputParser should handle this, but it doesn't
    ip.parse(gp, x, y, varargin{:});
else
    ip.parse(gp, x, y, [], varargin{:});
end
xt=ip.Results.xt;
z=ip.Results.z;
predcf=ip.Results.predcf;
nsamp=ip.Results.nsamp;
if isempty(xt)
    xt=x;
end

% get the parameters of Laplace approximation
[e, ~, ~, p] = gpla_eDMPS(gp_pak(gp), gp, x, y, 'z', z);
[f, LK, a, W] = deal(p.f, p.L, p.a, p.La2);

% [e, edata, eprior, p] = gpla_eDMPS(gp_pak(gp), gp, x, y, 'z', z);
% [f, L, W, p] = deal(p.f, p.L, p.La2, p.p);

% Calculate the covariance between training and prediction set
% notice the order xt,x to avoid transpose later
K_nf = gp_cov(gp,xt,x,predcf);

% Evaluate the mean
deriv = gp.lik.fh.llg(gp.lik, y, f, 'latent', z);
Eft = K_nf*deriv;

% Evaluate the covariance
kstarstar = gp_trcov(gp,xt,predcf);
V = LK'*W;
R = chol(eye(size(LK')) +  V*LK);
R2 = (R'\V)*K_nf';
Covft = kstarstar - K_nf*(W*K_nf') + R2'*R2;

% sample from the posterior
rr = randn(size(xt,1),nsamp);
sampft = bsxfun(@plus, Eft, chol(Covft,'lower')*rr);



end
