function lik = lik_countNumber(varargin)
%LIK_COUNTNUMBER   Create a Poisson likelihood structure for the DMPS
%                  inversion algorithm
%
%  Description
%    LIK = LIK_COUNTNUMBER('T', T) creates a Poisson likelihood structure
%                                  for the DMPS inversion algorithm
%
%    The likelihood is defined as follows:
%                  __ n
%      p(y|f, z) = || i=1 Poisson( y_i | z_i* sum( exp(f)*T ) )
%
%    where z is a vector of volumes of sampled air, f is the latent
%    value vector and T the transfer matrix. You must give the
%    vector z as an extra parameter to each function that requires
%    also y. For example, you can call gp_optim as follows:
%      gp_optimDMPS(gp, x, y, 'z', z)
%    If z is not given or it is empty, then z_i=1 is used.
%
%    Mandatory parameter for the likelihood is
%      T     - the transfer matrix
%
%  See also
%    GP_SET, LIK_*
%
%
%  References
%    Rasmussen, C. E. and Williams, C. K. I. (2006). Gaussian
%    Processes for Machine Learning. The MIT Press.
%
%    B. Mlgaard, J. Vanhatalo, P. Aalto, N.L. Prisle, and K. Hmeri
%    (manuscript). Notably improved inversion of Differential Mobility
%    Particle Sizer data obtained under conditions of fluctuating particle
%    number concentrations

% Copyright (c) 2015 Bjarke Molgaard
% Copyright (c) 2006-2010, 2014-2015 Jarno Vanhatalo
% Copyright (c) 2010 Aki Vehtari

% This software is distributed under the GNU General Public
% License (version 3 or later); please refer to the file
% License.txt, included with the software, for details.

ip=inputParser;
ip.FunctionName = 'LIK_COUNTNUMBER';
ip.addOptional('lik', [], @isstruct);
ip.addParameter('T',NaN, @(x) ismatrix(x));
ip.parse(varargin{:});
lik=ip.Results.lik;

if isempty(lik)
    init=true;
    lik.type = 'countNumber';
    lik.nondiagW = true;
    lik.Tr = ip.Results.T;
    if isnan(lik.Tr)
        error('LIK_COUNTNUMBER: you need to provide the transfer matrix')
    end
else
    if ~isfield(lik,'type') || ~isequal(lik.type,'countNumber')
        error('First argument does not seem to be a valid likelihood function structure')
    end
    init=false;
end

if init
    % Set the function handles to the subfunctions
    lik.fh.pak = @lik_countNumber_pak;
    lik.fh.unpak = @lik_countNumber_unpak;
    lik.fh.ll = @lik_countNumber_ll;
    lik.fh.llg = @lik_countNumber_llg;
    lik.fh.llg2 = @lik_countNumber_llg2;
    lik.fh.llg3 = @lik_countNumber_llg3;
    lik.fh.recappend = @lik_countNumber_recappend;
end

end

function [w,s,h] = lik_countNumber_pak(lik)
%LIK_COUNTNUMBER_PAK  Combine likelihood parameters into one vector.
%
%  Description
%    W = LIK_COUNTNUMBER_PAK(LIK) takes a likelihood structure LIK
%    and returns an empty verctor W. If Poisson likelihood had
%    hyperparameters this would combine them into a single row vector
%    W (see e.g. lik_negbin). This is a mandatory subfunction used
%    for example in energy and gradient computations.
%
%  See also
%    LIK_NEGBIN_UNPAK, GP_PAK

w = []; s = {}; h=[];
end


function [lik, w] = lik_countNumber_unpak(lik, w)
%LIK_COUNTNUMBER_UNPAK  Extract likelihood parameters from the vector.
%
%  Description
%    W = LIK_COUNTNUMBER_UNPAK(W, LIK) Doesn't do anything.
%
%    If Poisson likelihood had hyperparameters this would extract them
%    parameters from the vector W to the LIK structure. This is a
%    mandatory subfunction used for example in energy and gradient
%    computations.
%
%
%  See also
%    LIK_COUNTNUMBER_PAK, GP_UNPAK

lik=lik;
w=w;

end


function logLik = lik_countNumber_ll(lik, y, f, z)
%LIK_COUNTNUMBER_LL    Log likelihood
%
%  Description
%    E = LIK_COUNTNUMBER_LL(LIK, Y, F, Z) takes a likelihood 
%    structure LIK, particle counts Y, volumes Z, and
%    latent values F. Returns the log likelihood, log p(y|f,z).
%    This subfunction is needed when using Laplace approximation
%    or MCMC for inference with non-Gaussian likelihoods. This
%    subfunction is also used in information criteria (DIC, WAIC)
%    computations.
%
%  See also
%    LIK_COUNTNUMBER_LLG, LIK_COUNTNUMBER_LLG3, LIK_COUNTNUMBER_LLG2, GPLA_E

%n = size(y,1);
if numel(z)==0
    z=1;
end

lambda=(exp(f)'*lik.Tr)';
lambda = z.*lambda;
gamlny = gammaln(y+1);
logLik =  sum(-lambda + y.*log(lambda) - gamlny);
end


function deriv = lik_countNumber_llg(lik, y, f, param, z)
%LIK_COUNTNUMBER_LLG    Gradient of the log likelihood
%
%  Description
%    G = LIK_COUNTNUMBER_LLG(LIK, Y, F, PARAM) takes a likelihood
%    structure LIK, particle counts Y, volumes Z, and
%    latent values F. Returns the gradient of the log
%    likelihood with respect to PARAM. At the moment PARAM can be
%    'param' or 'latent'. This subfunction is needed when using
%    Laplace approximation or MCMC for inference with non-Gaussian
%    likelihoods.
%
%  See also
%    LIK_COUNTNUMBER_LL, LIK_COUNTNUMBER_LLG2, LIK_COUNTNUMBER_LLG3, GPLA_E

if numel(z)==0
    z=1;
end

switch param
    case 'latent'    % gradients with respect to the latent variables
        lambda=lik.Tr'*exp(f);
        Dlambda = bsxfun(@times,exp(f),lik.Tr);
        deriv = Dlambda*(y./lambda) - Dlambda*z;
    otherwise
        error('undefined param')
end

end


function g2 = lik_countNumber_llg2(lik, y, f, param, z)
%LIK_COUNTNUMBER_LLG2  Second gradients of the log likelihood 
%
%  Description
%    G2 = LIK_COUNTNUMBER_LLG2(LIK, Y, F, PARAM) takes a likelihood
%    structure LIK, particle counts Y, volumes Z, and
%    latent values F. Returns the Hessian of the log
%    likelihood with respect to PARAM. At the moment PARAM can be
%    only 'latent'. G2 is a vector with diagonal elements of the
%    Hessian matrix (off diagonals are zero). This subfunction
%    is needed when using Laplace approximation or EP for inference
%    with non-Gaussian likelihoods.

%
%  See also
%    LIK_COUNTNUMBER_LL, LIK_COUNTNUMBER_LLG, LIK_COUNTNUMBER_LLG3, GPLA_E

if numel(z)==0
    z=1;
end
switch param
    case 'latent'    % gradients with respect to the latent variables
        nf = length(f);
        lambda=lik.Tr'*exp(f);
        Dlambda = bsxfun(@times,exp(f),lik.Tr);
        yDl2 = y./lambda.^2;
        deriv = Dlambda*(y./lambda) - Dlambda*z;
        Dlambda_y=Dlambda.*(ones(nf,1)*yDl2');
        H=Dlambda_y*Dlambda';
        
        g2 = -H + sparse(1:nf,1:nf,deriv,nf,nf);
        
    otherwise
        error('undefined param')
end
end

function third_grad = lik_countNumber_llg3(lik, y, f, param, z)
%LIK_COUNTNUMBER_LLG3  Third gradients of the log likelihood
%
%  Description
%    G3 = LIK_COUNTNUMBER_LLG3(LIK, Y, F, PARAM) takes a likelihood
%    structure LIK, particle counts Y, volumes Z, and
%    latent values F and returns the third gradients of the
%    log likelihood with respect to PARAM. At the moment PARAM
%    can be only 'latent'. G3 is a cell array with third gradient matrices.
%    This subfunction is needed when using Laplace approximation
%    for inference with non-Gaussian likelihoods.
%
%  See also
%    LIK_COUNTNUMBER_LL, LIK_COUNTNUMBER_LLG, LIK_COUNTNUMBER_LLG2, GPLA_E, GPLA_G

if numel(z)==0
    z=1;
end
switch param
    case 'latent'          % gradients with respect to the latent variables
        nf = length(f);
        Tr = lik.Tr;
        lambda=Tr'*exp(f);
        Dlambda = bsxfun(@times,exp(f),Tr);
        deriv = - Dlambda*z + Dlambda*(y./lambda);
        
        third_grad = {};
        
        yDl2 = y./lambda.^2;
        H = sparse(nf,nf);
        for i1=1:length(y)
            H = H + Dlambda(:,i1)*Dlambda(:,i1)'*yDl2(i1);
        end
        D = sparse(eye(nf,nf));
        for h1=1:nf
            third_grad{h1} = sparse(nf,nf);
            col=find(Dlambda(h1,:));
            third_grad{h1} = third_grad{h1} + 2*Dlambda(:,col)*Dlambda(:,col)'.*Dlambda(h1,col)*(y(col)./lambda(col).^3);
            
            third_grad{h1} = third_grad{h1} -  D(:,h1)*H(h1,:) - H(:,h1)*D(h1,:);
                        
            rows=find(Dlambda(:,col));
            third_grad{h1} = third_grad{h1} - sparse(rows,rows,Dlambda(rows,col).*Dlambda(h1,col).*yDl2(col),nf,nf);
                        
            third_grad{h1}(h1,h1) = third_grad{h1}(h1,h1) + deriv(h1);
            
        end        
end
end


function reclik = lik_countNumber_recappend(reclik, ri, lik)
%RECAPPEND  Append the parameters to the record
%
%  Description
%    RECLIK = LIK_COUNTNUMBER_RECAPPEND(RECLIK, RI, LIK) takes a
%    likelihood record structure RECLIK, record index RI and
%    likelihood structure LIK with the current MCMC samples of
%    the parameters. Returns RECLIK which contains all the old
%    samples and the current samples from LIK. This subfunction
%    is needed when using MCMC sampling (gp_mc).
%
%  See also
%    GP_MC

if nargin == 2
    reclik.type = 'LIK_COUNTNUMBER';
    
    % Set the function handles
    reclik.fh.pak = @lik_countNumber_pak;
    reclik.fh.unpak = @lik_countNumber_unpak;
    reclik.fh.ll = @lik_countNumber_ll;
    reclik.fh.llg = @lik_countNumber_llg;
    reclik.fh.llg2 = @lik_countNumber_llg2;
    reclik.fh.llg3 = @lik_countNumber_llg3;
    reclik.fh.tiltedMoments = @lik_countNumber_tiltedMoments;
    reclik.fh.predy = @lik_countNumber_predy;
    reclik.fh.invlink = @lik_countNumber_invlink;
    reclik.fh.predprcty = @lik_countNumber_predprcty;
    reclik.fh.recappend = @lik_countNumber_recappend;
end
end
